/**h* NFCTools/nfc_pcsc_cards.c
 *
 * NAME
 *   nfc_pcsc_cards.c
 *
 * DESCRIPTION
 *   Manipulation of NFC contactless cards using PC/SC readers
 *
 * COPYRIGHT
 *   (c) 2011 PRO ACTIVE SAS - See LICENCE.txt for licence information
 *
 * AUTHOR
 *   Johann Dantant / PRO ACTIVE for www.springcard.com
 *
 * HISTORY
 *   JDA 17/01/2011 : first public release
 *
 **/
#include "nfc.h"
#include "nfc_card_i.h"

nfc_card_t nfc_card_open(const char *reader_name)
{
  LONG rc;
	unsigned long dummy_1, dummy_2, atrlen;
  struct _nfc_card_st *card = nfc_alloc(sizeof(struct _nfc_card_st));

	card->reader = reader_name;

  rc = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &card->hContext);
  if(rc != SCARD_S_SUCCESS)
  {
    printf("SCardEstablishContext : %08lX\n",rc);
	  goto failed;
  }

  rc = SCardConnect(card->hContext,
                    card->reader,
                    SCARD_SHARE_SHARED,
                    SCARD_PROTOCOL_T0|SCARD_PROTOCOL_T1,
                    &card->hCard,
                    &card->protocol);
  if(rc != SCARD_S_SUCCESS)
  {
	  if (!nfc_silent)
      printf("SCardConnect : %08lX\n",rc);
		SCardReleaseContext(card->hContext);
    nfc_free(card);
	  return NULL;
  }

  atrlen = sizeof(card->atr);
	rc = SCardStatus(card->hCard, NULL, NULL, &dummy_1, &dummy_2, card->atr, &atrlen);
  if(rc != SCARD_S_SUCCESS)
  {
	  if (!nfc_silent)
      printf("SCardStatus : %08lX\n",rc);
		SCardDisconnect(card->hCard, SCARD_EJECT_CARD);
		SCardReleaseContext(card->hContext);
    nfc_free(card);
	  return NULL;
  }
	card->atrlen = (unsigned char) atrlen;

  return card;

failed:
  exit(EXIT_FAILURE);
	return NULL;
}

void nfc_card_close(nfc_card_t card)
{
  if (card != NULL)
	{
    SCardDisconnect(card->hCard, SCARD_EJECT_CARD);
    SCardReleaseContext(card->hContext);
		nfc_free(card);
	}
}

SCARDHANDLE nfc_card_handle(nfc_card_t card)
{
  if (card == NULL)
	  return (SCARDHANDLE) 0;
  return card->hCard;
}

BOOL nfc_card_transmit(nfc_card_t card, unsigned char capdu[], unsigned long capdu_length, unsigned char rapdu[], unsigned long *rapdu_length)
{
  LONG rc;
	const SCARD_IO_REQUEST *pioSendPci;

  if (card == NULL)
	  return FALSE;

	switch (card->protocol)
	{
	  case SCARD_PROTOCOL_T0 : pioSendPci = SCARD_PCI_T0; break;
    case SCARD_PROTOCOL_T1 : pioSendPci = SCARD_PCI_T1; break;
		default                : pioSendPci = NULL;
	}

  rc = SCardTransmit(card->hCard,
	                   pioSendPci,
										 capdu,
										 capdu_length,
										 NULL,
										 rapdu,
										 rapdu_length);

  if (rc != SCARD_S_SUCCESS)
	{
	  if (!nfc_silent)
      printf("SCardTransmit error %08lX\n", rc);
		return FALSE;
	}

  return TRUE;
}


BOOL nfc_card_read_memory(nfc_card_t card, unsigned short address, unsigned char buffer[], unsigned char length)
{
  unsigned char capdu[5];
	unsigned char rapdu[256+2];
	unsigned long rapdu_length;

	if (card == NULL)
	  return FALSE;

	capdu[0] = 0xFF;
	capdu[1] = 0xB0;
	capdu[2] = (unsigned char) (address >> 8);
	capdu[3] = (unsigned char) (address);
	capdu[4] = length;


	rapdu_length = sizeof(rapdu);
	if (!nfc_card_transmit(card, capdu, sizeof(capdu), rapdu, &rapdu_length))
		return FALSE;

  if (rapdu_length < 2)
	{
    if (!nfc_silent)
		  printf("Card response is too short (no SW)\n");
    return FALSE;
	}

	if ((rapdu[rapdu_length-2] != 0x90) || (rapdu[rapdu_length-1] != 0x00))
	{
	  if (nfc_verbose)
		  printf("Card read error at address %d, SW=%02X%02X\n", address, rapdu[rapdu_length-2], rapdu[rapdu_length-1]);
		return FALSE;
	}

  if (rapdu_length-2 < length)
	{
	  if (nfc_verbose)
		  printf("Card read failed at address %u (length=%lu<%u)\n", address, rapdu_length-2, length);
    return FALSE;
	}

	if (buffer != NULL)
	  memcpy(buffer, rapdu, length);

  return TRUE;
}

BOOL nfc_card_write_memory(nfc_card_t card, unsigned short address, unsigned char buffer[], unsigned char length)
{
  unsigned char capdu[5+256];
	unsigned char rapdu[2];
	unsigned long rapdu_length;


	if (card == NULL)
    return FALSE;

	capdu[0] = 0xFF;
	capdu[1] = 0xD6;
	capdu[2] = (unsigned char) (address >> 8);
	capdu[3] = (unsigned char) (address);
	capdu[4] = length;
	memcpy(&capdu[5], buffer, length);

	rapdu_length = sizeof(rapdu);
	if (!nfc_card_transmit(card, capdu, 5 + length, rapdu, &rapdu_length))
    return FALSE;

  if (rapdu_length < 2)
	{
    if (!nfc_silent)
		  printf("Card response is too short (no SW)\n");
    return FALSE;
	}

	if ((rapdu[rapdu_length-2] != 0x90) || (rapdu[rapdu_length-1] != 0x00))
	{    
	  if (nfc_verbose)
		  printf("Card write error at address %d, SW=%02X%02X\n", address, rapdu[rapdu_length-2], rapdu[rapdu_length-1]);
		return FALSE;
	}

  return TRUE;
}

BOOL nfc_card_select_nfc_application(nfc_card_t card)
{
  static const unsigned char nfc_aid[] = NFC_NDEF_7816_AID;

  unsigned char capdu[5+256];
	unsigned char rapdu[256+2];
	unsigned long rapdu_length;

	if (card == NULL)
	  return FALSE;

	capdu[0] = 0x00;
	capdu[1] = 0xA4;
	capdu[2] = 0x04;
	capdu[3] = 0x00;
	capdu[4] = sizeof(nfc_aid);
	memcpy(&capdu[5], nfc_aid, sizeof(nfc_aid));

	rapdu_length = sizeof(rapdu);
	if (!nfc_card_transmit(card, capdu, 5 + sizeof(nfc_aid), rapdu, &rapdu_length))
		return FALSE;

  if (rapdu_length < 2)
	{
    if (!nfc_silent)
		  printf("Card response is too short (no SW)\n");
    return FALSE;
	}

	if ((rapdu[rapdu_length-2] != 0x90) || (rapdu[rapdu_length-1] != 0x00))
	{
	  if (nfc_verbose)
		  printf("Card select NDEF application error, SW=%02X%02X\n", rapdu[rapdu_length-2], rapdu[rapdu_length-1]);
		return FALSE;
	}

  return TRUE;
}

BOOL nfc_card_select_file(nfc_card_t card, unsigned short file_id)
{
  unsigned char capdu[5+2];
	unsigned char rapdu[256+2];
	unsigned long rapdu_length;

	if (card == NULL)
	  return FALSE;

	capdu[0] = 0x00;
	capdu[1] = 0xA4;
	capdu[2] = 0x00;
	capdu[3] = 0x0C;
	capdu[4] = 2;
	capdu[5] = (unsigned char) (file_id >> 8);
	capdu[6] = (unsigned char) (file_id);

	rapdu_length = sizeof(rapdu);
	if (!nfc_card_transmit(card, capdu, 5 + 2, rapdu, &rapdu_length))
		return FALSE;

  if (rapdu_length < 2)
	{
    if (!nfc_silent)
		  printf("Card response is too short (no SW)\n");
    return FALSE;
	}

	if ((rapdu[rapdu_length-2] != 0x90) || (rapdu[rapdu_length-1] != 0x00))
	{
	  if (nfc_verbose)
		  printf("Card select file %02X error, SW=%02X%02X\n", file_id, rapdu[rapdu_length-2], rapdu[rapdu_length-1]);
		return FALSE;
	}

  return TRUE;
}

BOOL nfc_card_read_binary(nfc_card_t card, unsigned short offset, unsigned char buffer[], unsigned char length)
{
  unsigned char capdu[5];
	unsigned char rapdu[256+2];
	unsigned long rapdu_length;

	if (card == NULL)
	  return FALSE;

	capdu[0] = 0x00;
	capdu[1] = 0xB0;
	capdu[2] = (unsigned char) (offset >> 8);
	capdu[3] = (unsigned char) (offset);
	capdu[4] = length;


	rapdu_length = sizeof(rapdu);
	if (!nfc_card_transmit(card, capdu, sizeof(capdu), rapdu, &rapdu_length))
		return FALSE;

  if (rapdu_length < 2)
	{
    if (!nfc_silent)
		  printf("Card response is too short (no SW)\n");
    return FALSE;
	}

	if ((rapdu[rapdu_length-2] != 0x90) || (rapdu[rapdu_length-1] != 0x00))
	{
	  if (nfc_verbose)
		  printf("Card read error at offset %d, SW=%02X%02X\n", offset, rapdu[rapdu_length-2], rapdu[rapdu_length-1]);
		return FALSE;
	}

  if (rapdu_length-2 < length)
	{
	  if (nfc_verbose)
		  printf("Card read failed at offset %u (length=%lu<%u)\n", offset, rapdu_length-2, length);
    return FALSE;
	}

	if (buffer != NULL)
	  memcpy(buffer, rapdu, length);

  return TRUE;
}

BOOL nfc_card_update_binary(nfc_card_t card, unsigned short offset, unsigned char buffer[], unsigned char length)
{
  unsigned char capdu[5+256];
	unsigned char rapdu[2];
	unsigned long rapdu_length;

	if (card == NULL)
	  return FALSE;

	capdu[0] = 0x00;
	capdu[1] = 0xD6;
	capdu[2] = (unsigned char) (offset >> 8);
	capdu[3] = (unsigned char) (offset);
	capdu[4] = length;
	memcpy(&capdu[5], buffer, length);

	rapdu_length = sizeof(rapdu);
	if (!nfc_card_transmit(card, capdu, 5 + length, rapdu, &rapdu_length))
		return FALSE;

  if (rapdu_length < 2)
	{
    if (!nfc_silent)
		  printf("Card response is too short (no SW)\n");
    return FALSE;
	}

	if ((rapdu[rapdu_length-2] != 0x90) || (rapdu[rapdu_length-1] != 0x00))
	{
	  if (nfc_verbose)
		  printf("Card write error at offset %d, SW=%02X%02X\n", offset, rapdu[rapdu_length-2], rapdu[rapdu_length-1]);
		return FALSE;
	}

  return TRUE;
}
