/**h* NFCTools/nfc_type_1.c
 *
 * NAME
 *   nfc_type_1.c
 *
 * DESCRIPTION
 *   Support of NFC Type 1 Tag
 *
 * COPYRIGHT
 *   (c) 2011 PRO ACTIVE SAS - See LICENCE.txt for licence information
 *
 * AUTHOR
 *   Johann Dantant / PRO ACTIVE for www.springcard.com
 *
 * HISTORY
 *   JDA 17/01/2011 : first public release
 *
 **/
#include "nfc.h"
#include "nfc_card_i.h"

BOOL nfc_card_topaz_recognize(nfc_card_t card)
{
	unsigned char buffer[120];
	unsigned char i;

	if (card == NULL)
	  return FALSE;

  if (nfc_verbose)
		printf("\tthis is a Topaz chip\n");

	card->type = CARD_INNOVISION_TOPAZ;
	card->capacity = 92;

	if (!nfc_card_read_memory(card, 0, buffer, sizeof(buffer)))
	{
		if (nfc_verbose)
			printf("\tfailed to read the card memory\n");
		return FALSE;
	}

  /* Check lock bits */
	card->writable = TRUE;
	if ((buffer[112] & 0xFE) || (buffer[113] & 0x1F))
	{
		card->writable = FALSE;
		if (nfc_verbose)
			printf("\tlock bits are set\n");
	}

	/* Analyze the CC - it must either be empty, or a valid CC */
	if ((buffer[8] == NFC_FORUM_MAGIC_NUMBER) && (buffer[9] != 0x00) && (buffer[10] != 0x00))
	{
	  size_t size = 8 * (buffer[10] + 1);

		card->prepared = TRUE;
    card->capacity = size - 8 - 16;

		if (nfc_verbose)
			printf("\tCC are valid, memory size=%dB, capacity=%dB\n", size, card->capacity);

		if (buffer[11] & 0x0F)
		{
			card->writable = FALSE;
			if (nfc_verbose)
        printf("\tCC says the card is not writable\n");
		}
	} else
	if ((buffer[12] == 0x00) || (buffer[13] == 0x00) || (buffer[14] == 0x00) || (buffer[15] == 0x00))
	{
		if (nfc_verbose)
      printf("\tCC are blank\n");
	} else
	{
		if (nfc_verbose)
      printf("\tCC are invalid\n");
	}

	/* Is card empty ? */
	card->empty = TRUE;
	for (i=8; i<96+8; i++)
	{
		if (buffer[i] != 0x00)
		{
			if (nfc_verbose)
				printf("\tcard is not empty\n");
			card->empty = FALSE;
			break;
		}
	}

  return TRUE;
}

BOOL nfc_card_topaz_prepare(nfc_card_t card)
{
  BYTE buffer[4];
	int i;
	unsigned short addr = 8;

  if ((card == NULL) || (card->type != CARD_INNOVISION_TOPAZ))
    return FALSE;

	buffer[0] = NFC_FORUM_MAGIC_NUMBER;
	buffer[1] = NFC_FORUM_VERSION_NUMBER;
	buffer[2] = (120 / 8) - 1;
	buffer[3] = 0x00;

	if (nfc_verbose)
		printf("\twriting the CC on block 1\n");

  for (i=0; i<4; i++)
	{
		if (!nfc_card_write_memory(card, addr++, &buffer[i], 1))
		{
			if (!nfc_silent)
				printf("Failed to write the CC%d into the Topaz card\n", i);
			return FALSE;
		}
	}

	return TRUE;
}

BOOL nfc_card_topaz_lock(nfc_card_t card)
{
  BYTE lock = 0xFF;
	int i;
	unsigned short addr = 112;

  if ((card == NULL) || (card->type != CARD_INNOVISION_TOPAZ))
    return FALSE;

  for (i=0; i<2; i++)
	{
	  /* Normal write */
		if (!nfc_card_write_memory(card, addr, &lock, 1))
		{
		  /* Write not-erase function (specific for Topaz...) */
			if (!nfc_card_write_memory(card, (unsigned short) (0x8000 | addr), &lock, 1))
			{
				/* NOTE : it seems that the Topaz doesn't confirm the writing, */
				/*        so we read back the data to check...                 */				
				BYTE readback;

				if ((nfc_card_read_memory(card, addr, &readback, 1) && (readback == lock)))
				  continue;

				if (!nfc_silent)
					printf("Failed to write the Lock%d into the Topaz card\n", i);
				return FALSE;
			}
		}

		addr++;
	}

	return TRUE;
}

BOOL nfc_card_topaz_write_content(nfc_card_t card, unsigned char *content, size_t content_length)
{
	size_t i;
	unsigned short addr = 12;

  if ((card == NULL) || (card->type != CARD_INNOVISION_TOPAZ))
    return FALSE;

  for (i=0; i<content_length; i++)
	{
	  if (i>card->capacity)
		  break;
		if (!nfc_card_write_memory(card, addr++, &content[i], 1))
		{
			if (!nfc_silent)
				printf("Failed to write at address %d into the Topaz card\n", i);
			return FALSE;
		}
	}

	return TRUE;
}
