﻿using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using System.Globalization;
using System.Linq;
using System.Text;
using System.Runtime.InteropServices;
using System.Diagnostics;
using Microsoft.Win32;

namespace SpringCard.LibCs
{
	public class IniFile
	{		
		private struct Entry
    	{
        	public string Section;
	        public string Name;
	        public string Value;
	    }
		
		private List<string> Sections;		
		private List<Entry> Entries;
		
		private bool _ReadOnly = false;
		private bool _AutoSave = true;
        private string _FileName;
        
        public IniFile(string FileName)
        {
        	_ReadOnly = false;
        	_AutoSave = true;
            _FileName = FileName;
            Load();
        }

        public IniFile(string FileName, bool AutoSave)
        {
        	_ReadOnly = false;
        	_AutoSave = AutoSave;
            _FileName = FileName;
            Load();
        }

        public IniFile(string FileName, bool AutoSave, bool ReadOnly)
        {
        	_ReadOnly = ReadOnly;
        	_AutoSave = AutoSave;
            _FileName = FileName;
            Load();
        }
        
		public bool Load()
    	{
			bool Result = true;
			
        	TextReader iniFile = null;
	        string strLine = null;
    	    string strSection = null;
    	    
    	    Sections = new List<string>();
    	    Entries = new List<Entry>();    	    

	        if (File.Exists(_FileName))
        	{
            	try
	            {
    	            iniFile = new StreamReader(_FileName);

                	strLine = iniFile.ReadLine();

                	while (strLine != null)
                	{
                    	strLine = strLine.Trim();
                    	
                    	if (strLine.Contains(";"))
                    		strLine = strLine.Substring(0, strLine.IndexOf(';'));

	                    if (strLine != "")
    	                {
    	                    if (strLine.StartsWith("[") && strLine.EndsWith("]"))
    	                    {
    	                    	/* Start of section */
    	                        strSection = strLine.Substring(1, strLine.Length - 2);
    	                        
    	                        /* Add to sections */
    	                        Sections.Add(strSection);
    	                    }
    	                    else
    	                    {
    	                        Entry e = new Entry();
    	                        e.Section = strSection;
    	                        
    	                    	string[] t = strLine.Split(new char[] { '=' }, 2);    	                        
    	                        e.Name = t[0];
    	                        e.Value = null;
    	                        if (t.Length > 1)
    	                            e.Value = t[1];

    	                        Entries.Add(e);
    	                    }
    	                }

    	                strLine = iniFile.ReadLine();
    	            }

    	        }
    	        catch (Exception ex)
    	        {
    	        	Console.WriteLine(ex.Message);
    	            Result = false;
    	        }
    	        finally
    	        {
    	            if (iniFile != null)
    	                iniFile.Close();
    	        }
    	    }
    	    else
    	    {
    	    	Console.WriteLine(_FileName + " not found!");
				Result = false;
    	    }
    	    
    	    return Result;
	    }
		
		public bool SaveTo(string FileName)
		{
			string oldFileName = _FileName;
			_FileName = FileName;
			if (!Save())
			{
				_FileName = oldFileName;
				return false;
			}
			return true;
		}
		
		public bool Save()
		{
			if (_ReadOnly)
				return false;				
			
	        string strToSave = "";	        

            foreach (Entry entry in Entries)
	        {
            	if ((entry.Section == null) || (entry.Section == ""))
	            {
	            	if (entry.Value == null)
	            		strToSave += entry.Name;
	            	else
	            		strToSave += entry.Name + "=" + entry.Value;

	                strToSave += "\r\n";
	            }
	        }	        
	        
    	    foreach (string section in Sections)
        	{
            	strToSave += "[" + section + "]" + "\r\n";

	            foreach (Entry entry in Entries)
    	        {
    	            if (entry.Section == section)
    	            {
    	            	if (entry.Value == null)
    	            		strToSave += entry.Name;
    	            	else
    	            		strToSave += entry.Name + "=" + entry.Value;

    	                strToSave += "\r\n";
    	            }
    	        }

    	        strToSave += "\r\n";
        	}

        	try
        	{
            	TextWriter tw = new StreamWriter(_FileName);
	            tw.Write(strToSave);
    	        tw.Close();
    	        return true;
        	}
        	catch (Exception ex)
        	{
            	Console.WriteLine(ex.Message);
            	return false;
        	}			
		}

		public string ReadString(string Section, string Name, string Default)
    	{
			Console.WriteLine("INI: Read [" + Section + "] " + Name);
			
            foreach (Entry entry in Entries)
	        {
	            if (entry.Section == Section)
	            {
	            	if (entry.Name == Name)
	            	{
	            		Console.WriteLine("INI: Found value '" + entry.Value + "'");
	            		return entry.Value;
	            	}
            	}
	        }
            
            return Default;
    	}		
        
		public string ReadString(string Section, string Name)
        {
			return ReadString(Section, Name, "");
        }
		
		public int ReadInteger(string Section, string Name, int Default)
		{
			string s = ReadString(Section, Name);
			int r;
			if (int.TryParse(s, out r))
				return r;
			return Default;
		}
		
		public int ReadInteger(string Section, string Name)
		{
			return ReadInteger(Section, Name, 0);
		}
		
		public bool ReadBoolean(string Section, string Name)
		{
			int i;
			i = ReadInteger(Section, Name, 0);
			return (i != 0) ? true : false;		
		}
		
		public bool Erase(string Section, string Name)
		{
			if (_ReadOnly)
				return false;				
			
			for (int i=0; i<Entries.Count; i++)
			{
				Entry oldEntry = Entries[i];
				if (oldEntry.Section == Section)
				{
					if (oldEntry.Name == Name)
					{
						Entries.Remove(oldEntry);
						break;
					}
				}
			}			
			
	        if (_AutoSave)
	        	return Save();
	        
	        return true;			
		}
		
		public bool WriteString(string Section, string Name, string Value)
	    {
			if (_ReadOnly)
				return false;				
			
			bool isNewEntry = true;
			bool isNewSection = true;
			
			Entry newEntry = new Entry();
			newEntry.Section = Section;
			newEntry.Name = Name;
			newEntry.Value = Value;
			
			if ((Section != null) && (Section != ""))
			{
				for (int i=0; i<Sections.Count; i++)
				{
					if (Sections[i] == Section)
					{
						isNewSection = false;
						break;
					}
				}
			} else
			{
				isNewSection = false;
			}			
			
			for (int i=0; i<Entries.Count; i++)
			{
				Entry oldEntry = Entries[i];
				if (oldEntry.Section == Section)
				{
					if (oldEntry.Name == Name)
					{
						Entries[i] = newEntry;
						isNewEntry = false;
						break;
					}
				}
			}

			if (isNewSection)
			{
				Sections.Add(Section);
			}
			
			if (isNewEntry)
			{
				Entries.Add(newEntry);
			}
			
			
	        if (_AutoSave)
	        	return Save();
	        
	        return true;
	    }	
		
        public bool WriteInteger(string Section, string Name, int Value)
        {
        	return WriteString(Section, Name, String.Format("{0}", Value));
        }
		
		public bool WriteBoolean(string Section, string Name, bool value)
		{
			int i = value ? 1 : 0;
			return WriteInteger(Section, Name, i);		
		}
		
		public string[] ReadSection(string Section)
		{
			return ReadSection(Section, false);
		}
		
 		public string[] ReadSection(string Section, bool withAssociatedValues)
    	{
			Console.WriteLine("INI: List [" + Section + "]");
			
			List<string> Names = new List<string>();
			 			
            foreach (Entry entry in Entries)
	        {
	            if (entry.Section == Section)
	            {           		
            		if (withAssociatedValues)
            		{
	            		Console.WriteLine("INI: Found " + entry.Name + "=" + entry.Value);            	            			
	            		Names.Add(entry.Name + "=" + entry.Value);
            			
            		} else
            		{
            			Console.WriteLine("INI: Found " + entry.Name);
	            		Names.Add(entry.Name);
            		}
            	}
	        }
            
            return Names.ToArray();
    	}
	}
 
}
