/*
 * NAME
 *   ms_ccid_enable.c
 *
 * DESCRIPTION
 *   Command line utility to activate the support of Vendor Escape commands
 *   for CCID readers connected through Microsoft CCID driver.
 *   (i.e. to all SCardControl function calls with direct exchange with the
 *   reader) 
 *
 * COPYRIGHT
 *   Copyright (c) 2010 Johann Dantant / Pro-Active
 *
 **/
#include <windows.h>
#pragma comment(lib,"winscard.lib")

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static const char hklm_system_css_enum_usb[] = "HKLM\\SYSTEM\\CurrentControlSet\\Enum\\USB";
static const char wudf_ccid_driver[]         = "WUDFUsbccidDriver";
static const char device_parameters[]        = "Device Parameters";
static const char escape_command_enable[]    = "EscapeCommandEnable";
static const char product_id_mask[]          = "Vid_1c34&Pid_71";

size_t strlcpy(char *dst, const char *src, size_t dstsize);
size_t strlcat(char *dst, const char *src, size_t dstsize);

void REGFILE_WriteDWord(const char *kname, const char *name, unsigned long value);
int REGFILE_ReadSections(const char *kname, int num, char *name, int len);

void process_pid_vid_serno(const char *product_pid_vid, const char *serno)
{
  char product_full_key[256];
	char parameters_key[256];

	strlcpy(product_full_key, hklm_system_css_enum_usb, sizeof(product_full_key));
	strlcat(product_full_key, "\\", sizeof(product_full_key));
	strlcat(product_full_key, product_pid_vid, sizeof(product_full_key));
	strlcat(product_full_key, "\\", sizeof(product_full_key));
	strlcat(product_full_key, serno, sizeof(product_full_key));

  strlcpy(parameters_key, product_full_key, sizeof(parameters_key));
	strlcat(parameters_key, "\\", sizeof(parameters_key));
	strlcat(parameters_key, device_parameters, sizeof(parameters_key));

  REGFILE_WriteDWord(parameters_key, escape_command_enable, 1);

  strlcat(parameters_key, "\\", sizeof(parameters_key));
	strlcat(parameters_key, wudf_ccid_driver, sizeof(parameters_key));
	REGFILE_WriteDWord(parameters_key, escape_command_enable, 1);
}

void process_pid_vid(const char *product_pid_vid)
{
  int i, r = 1;
  char product_pid_vid_key[256];
	char product_serno[256];


	strlcpy(product_pid_vid_key, hklm_system_css_enum_usb, sizeof(product_pid_vid_key));
	strlcat(product_pid_vid_key, "\\", sizeof(product_pid_vid_key));
	strlcat(product_pid_vid_key, product_pid_vid, sizeof(product_pid_vid_key));

	for (i=0; r==1; i++)
	{
    r = REGFILE_ReadSections(product_pid_vid_key, i, product_serno, sizeof(product_serno));

		if ((r == 1) && (strlen(product_serno) == 8))
		{
      printf("%s\n", product_serno);

		  process_pid_vid_serno(product_pid_vid, product_serno);
		}

	}

}

int main(int argc, char **argv)
{
  int i, r = 1;
  char product_pid_vid[256];

	for (i=0; r==1; i++)
	{
    r = REGFILE_ReadSections(hklm_system_css_enum_usb, i, product_pid_vid, sizeof(product_pid_vid));

		if (r == 1)
		{
		  if (!_strnicmp(product_pid_vid, product_id_mask, strlen(product_id_mask)))
			{
        printf("%s\n", product_pid_vid);

				process_pid_vid(product_pid_vid);
			}
		}

	}


  return EXIT_SUCCESS;
}





/*
 * Copy src to string dst of size siz.  At most siz-1 characters
 * will be copied.  Always NUL terminates (unless siz == 0).
 * Returns strlen(src); if retval >= siz, truncation occurred.
*/
size_t strlcpy(char *dst, const char *src, size_t dstsize)
{
	register char *d = dst;
	register const char *s = src;
	register size_t n = dstsize;
	/* Copy as many bytes as will fit */
	if (n != 0 && --n != 0) {
	do {
		if ((*d++ = *s++) == 0)
			break;
		} while (--n != 0);
	}

	/* Not enough room in dst, add NUL and traverse rest of src */
	if (n == 0) {
		if (dstsize != 0)
			*d = '\0';      /* NUL-terminate dst */
		while (*s++);
	}

	return(s - src - 1);    /* count does not include NUL */
}

size_t strlcat(char *dst, const char *src, size_t dstsize)
{
	char *df = dst;
  size_t left = dstsize;
  size_t l1;
  size_t l2 = strlen(src);
  size_t copied;

	while (left-- != 0 && *df != '\0')
  	df++;
  l1 = df - dst;
  if (dstsize == l1)
  	return (l1 + l2);

  copied = l1 + l2 >= dstsize ? dstsize - l1 - 1 : l2;
  (void) memcpy(dst + l1, src, copied);
  dst[l1+copied] = '\0';
  return (l1 + l2);
}


typedef struct
{
  const char *prefix;
  HKEY       hkey;
} HKEY_DEF_ST;

static const HKEY_DEF_ST hkey_defs[] = 
{
  { "HKLM",               HKEY_LOCAL_MACHINE },
  { "HKEY_LOCAL_MACHINE", HKEY_LOCAL_MACHINE },
  { "HKCU",               HKEY_CURRENT_USER  },
  { "HKEY_CURRENT_USER",  HKEY_CURRENT_USER  },
  { "",                   0                  }
};

BOOL translate_key_name(const char *kname, HKEY *k_hkey, const char **k_rname)
{
  int i;

  for (i=0; strlen(hkey_defs[i].prefix); i++)
  {
    if (!_strnicmp(kname, hkey_defs[i].prefix, strlen(hkey_defs[i].prefix)))
    {
      *k_hkey  = hkey_defs[i].hkey;
      *k_rname = kname + strlen(hkey_defs[i].prefix) + 1;
      return TRUE;
    }
  }

  return FALSE;
}

int REGFILE_ReadSections(const char *kname, int num, char *name, int len)
{
  HKEY        k_hkey;
  const char *k_rname;

  HKEY        hKey;
  DWORD       count, size;

  if (!translate_key_name(kname, &k_hkey, &k_rname)) return -1;

  if (RegOpenKeyEx(k_hkey, k_rname, 0, KEY_READ|KEY_EXECUTE, &hKey) != ERROR_SUCCESS) return -1;

  if (RegQueryInfoKey(hKey, NULL, NULL, NULL, &count, NULL, NULL, NULL, NULL, NULL, NULL, NULL) != ERROR_SUCCESS)
  {
    RegCloseKey(hKey);
    return -1;
  }

  if ((DWORD) num >= count) return 0;

  size = len;
  if (RegEnumKeyEx(hKey, num, name, &size, NULL, NULL, NULL, NULL) != ERROR_SUCCESS)
  {
    RegCloseKey(hKey);
    return -1;
  }


  RegCloseKey(hKey);
  return 1;
}

void REGFILE_WriteDWord(const char *kname, const char *name, unsigned long value)
{
  HKEY        k_hkey;
  const char *k_rname;

  HKEY        hKey;

  if (!translate_key_name(kname, &k_hkey, &k_rname)) return;

  if (RegOpenKeyEx(k_hkey, k_rname, 0, KEY_SET_VALUE|KEY_EXECUTE, &hKey) != ERROR_SUCCESS) return;

  RegSetValueEx(hKey, name, 0, REG_DWORD, (LPBYTE) &value, sizeof(unsigned long));

  RegCloseKey(hKey);
}
