/**h* PCSCSneak/pcscsneak.c
 *
 * NAME
 *   pcscsneak.c :: PC/SC Sneaker (C implementation)
 *
 * DESCRIPTION
 *   This sample program shows how to use SCardGetStatusChange to track smartcard events
 *   (add/remove reader to the system, insert/remove a card in a reader).
 *   Any changes are timestamped and recorded into a LOG file.
 *
 * COPYRIGHT
 *   Copyright (c) 2012 Johann Dantant / Pro-Active SAS
 *
 **/
#define _CRT_SECURE_NO_WARNINGS
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <winscard.h>
#include "pcscsneak.h"

static const char *logfile = "C:\\SNEAKERS.LOG";


/* Visual Studio : link to Windows' winscard library */
#ifdef WIN32
#pragma comment(lib,"winscard.lib")
#endif

int main(int argc, char **argv)
{
	SCARDCONTEXT      hContext;
  DWORD  dwReaders = 0, dwReadersOld = 0;
	LPSTR  szReaders = NULL; 
  _TIMESTAMP_ST old_timestamp, new_timestamp;
  SCARD_READERSTATE *rgscState = NULL;
  DWORD  dwGlobalState = SCARD_STATE_UNAWARE;
	BOOL   bListPrepared = FALSE;
  BOOL   bFirstLoopDone = FALSE;
  DWORD  i;
	LONG   rc;

  /*
   * Welcome message, check parameters
   * ---------------------------------
   */

  report_init("C:\\SNEAKERS.LOG");
  init_timestamps();
  report_single("Program starting");

	/*
   * Get a handle to the PC/SC resource manager
   * ------------------------------------------
   */
	rc = SCardEstablishContext(SCARD_SCOPE_SYSTEM, NULL, NULL, &hContext);
	if(rc != SCARD_S_SUCCESS)
	{
		report_single("SCardEstablishContext :%lXh",rc);
		return EXIT_FAILURE;
	}

	/*
   * Infinite loop, we'll exit only when killed by Ctrl+C
   * ----------------------------------------------------
   */

  get_timestamp(&new_timestamp);

	for (;;)
	{
  	/*
     * Get the list of readers
     * -----------------------
  	 */
    LPTSTR pReader;
    DWORD  dwReadersSz;
    BOOL   bGlobalEventFired = FALSE;

    if (bListPrepared)
    {
      /* Free the list of readers  */
      if (szReaders != NULL)
      {
        SCardFreeMemory(hContext, szReaders);
        szReaders = NULL;
      }

      /* Allocate and populate the new list */
      dwReadersSz = SCARD_AUTOALLOCATE;
		  rc = SCardListReaders(hContext,
                            NULL,                /* Any group */
                            (LPTSTR) &szReaders, /* Diabolic cast for buffer auto-allocation */
                            &dwReadersSz);       /* Beg for auto-allocation */

      memcpy(&old_timestamp, &new_timestamp, sizeof(_TIMESTAMP_ST));
      get_timestamp(&new_timestamp);

		  if (rc == SCARD_E_NO_READERS_AVAILABLE)
		  {
		    /* No reader at all */
			  dwReaders = 0;
        szReaders = NULL;
		  } else
      if (rc != SCARD_S_SUCCESS)
		  {
        report_single("SCardListReaders error %lXh",rc);
			  continue;
		  }
    
		  /* Count the readers */
      if (szReaders != NULL)
      {
		    pReader = szReaders;
		    for (i=0; ; i++)
		    {
          /* End of multi-string array */
			    if (*pReader == '\0')
          {
            dwReaders = i;
				    break;
          }
          /* Jump to next entry in multi-string array */
			    pReader += strlen(pReader) + 1;
        }
      }

      if (dwReadersOld != dwReaders)
		  {
        if (!bFirstLoopDone)
        {
          report_single("%d readers found", dwReaders);
        } else
        {
          if (dwReadersOld > dwReaders)
          {
            report_single("%d readers removed (%d --> %d)", dwReadersOld - dwReaders, dwReadersOld, dwReaders);
          } else
          {
            report_single("%d readers added (%d --> %d)", dwReaders - dwReadersOld, dwReadersOld, dwReaders);
          }
			  }

        dwReadersOld = dwReaders;
		  }
    }

    /* Allocate the list of status */
    /* --------------------------- */

    if (rgscState != NULL)
    {
      free(rgscState);
      rgscState = NULL;
    }

    rgscState = calloc(1 + dwReaders, sizeof(SCARD_READERSTATE));
    if (rgscState == NULL)
    {
      report_single("Out of memory");
      return EXIT_FAILURE;
    }

    /* The first entry is to monitor the readers insertion/removal */
    rgscState[0].szReader = "\\\\?PNP?\\NOTIFICATION";
    rgscState[0].dwCurrentState = dwGlobalState;

    /* Put the name of readers in the list */
		pReader = szReaders;
		for (i=0; i<dwReaders; i++)
		{
		  rgscState[1 + i].szReader = pReader;
      rgscState[i].dwEventState = SCARD_STATE_UNAWARE;
      pReader += strlen(pReader) + 1;
    }

		/*
     * Interesting part of the job : call SCardGetStatusChange to monitor all changes
     * taking place in the PC/SC reader(s)
     */
    while (!bGlobalEventFired)
    {
		  rc = SCardGetStatusChange(hContext,
			                          INFINITE,
			                          rgscState,
			                          1 + dwReaders);	
		  
		  if (rc != SCARD_S_SUCCESS)
      {
        report_single("SCardGetStatusChange error %lXh",rc);

        for (i=1; i<=dwReaders; i++)
          rgscState[i].dwEventState = SCARD_STATE_UNAWARE;

        if (rc == SCARD_F_INTERNAL_ERROR)
        {
          /* Oups ? Restard from scratch */
          break;
        }
      } else
      {
        report_multi_begin();

		    for (i=0; i<=dwReaders; i++)
		    {
          if (i == 0)
          {
            if (rgscState[i].dwEventState & SCARD_STATE_CHANGED)
            {
              /* The list of readers has changed */
              /* ------------------------------- */

              report_multi_add("List of readers has changed");

              dwGlobalState = rgscState[i].dwEventState;
              bGlobalEventFired = TRUE;
            }

          } else
          {
            BOOL bCardInserted = FALSE;

            if (rgscState[i].dwEventState & SCARD_STATE_CHANGED)
            {
              /* Something happenned on the reader */
              /* -------------------------------- */

              if ( (rgscState[i].dwEventState & SCARD_STATE_PRESENT)
               && !(rgscState[i].dwCurrentState & SCARD_STATE_PRESENT) )
                bCardInserted = TRUE;

            } else
            {
              /* Nothing new, don't display anything for this reader */
              /* (unless we're in the first run)                     */
              if (bFirstLoopDone)
                continue;
            }

            /* Remember new current state for next loop */
            rgscState[i].dwCurrentState = rgscState[i].dwEventState;

			      /* Reader's name */
            report_multi_add("- %s", rgscState[i].szReader);

            if (rgscState[i].dwEventState & SCARD_STATE_UNKNOWN)
              report_multi_add("  UNKNOWN");
            if (rgscState[i].dwEventState & SCARD_STATE_UNAVAILABLE)
              report_multi_add("  UNAVAILABLE");
            if (rgscState[i].dwEventState & SCARD_STATE_EMPTY)
              report_multi_add("  EMPTY");
            if (rgscState[i].dwEventState & SCARD_STATE_PRESENT)
              report_multi_add("  PRESENT");
            if (rgscState[i].dwEventState & SCARD_STATE_MUTE)
              report_multi_add("  MUTE");
            if (rgscState[i].dwEventState & SCARD_STATE_INUSE)
              report_multi_add("  INUSE");
            if (rgscState[i].dwEventState & SCARD_STATE_EXCLUSIVE)
              report_multi_add("  EXCLUSIVE");
            if (rgscState[i].dwEventState & SCARD_STATE_UNPOWERED)
              report_multi_add("  UNPOWERED");

            if (bCardInserted)
            {
              /*
               * Display card's ATR
               * ------------------
               */

              if (rgscState[i].cbAtr)
				        report_multi_add_buffer("  ATR=", rgscState[i].rgbAtr, rgscState[i].cbAtr);
            }
          }
        }

        report_multi_end();
		  }

      bListPrepared = TRUE;
      if (dwReaders)
      {
        /* Not the first run anymore */
        bFirstLoopDone = TRUE;
      }
    }
	}


	rc = SCardReleaseContext(hContext);

  /* Never go here (Ctrl+C kill us before !) */
  free(rgscState);
  return EXIT_SUCCESS;
}


